/********************************************************************************
 *  FARSA - Total99                                                             *
 *  Copyright (C) 2012-2013 Gianluca Massera <emmegian@yahoo.it>                *
 *                                                                              *
 *  This program is free software; you can redistribute it and/or modify        *
 *  it under the terms of the GNU General Public License as published by        *
 *  the Free Software Foundation; either version 2 of the License, or           *
 *  (at your option) any later version.                                         *
 *                                                                              *
 *  This program is distributed in the hope that it will be useful,             *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *  GNU General Public License for more details.                                *
 *                                                                              *
 *  You should have received a copy of the GNU General Public License           *
 *  along with this program; if not, write to the Free Software                 *
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA  *
 ********************************************************************************/

#ifndef CARTPOLEEXPERIMENT_H
#define CARTPOLEEXPERIMENT_H

#include "farsaplugin.h"
#include "evorobotexperiment.h"
#include "evonet.h"
#include "wvector.h"
#include <QVector>

/**
 * \brief An experiment in which a khepera robot has to discriminate between an
 *        object and the arena walls
 *
 * The resources used by this experiment are the same as the ones of the parent
 * class (EvoRobotExperiment)
 *
 * The parameters for this experiment are:
 * 	- distanceThreshold: the distance from the object below which the robot
 * 	                     is rewarded. This is the distance of the nearest
 * 	                     points of the robot and the object
 * 	- playgroundWidth: the width of the part of the arena surrounded by
 * 	                   walls (the playground is the area where the robot can
 * 	                   move)
 * 	- playgroundHeight: the height of the part of the arena surrounded by
 * 	                    walls (the playground is the area where the robot
 * 	                    can move)
 * 	- minObjectDistanceFromWall: the minimum distance from the walls at
 * 	                             which the object is placed. The default is
 * 	                             0.05
 * 	- minInitialRobotDistanceFromObject: the minimum distance from the
 * 	                                     object at which the robot is
 * 	                                     initialized. The default is 0.1
 * 	- minInitialRobotDistanceFromWall: the minimum distance from the walls
 * 	                                   at which the robot is initialized.
 * 	                                   The default is 0.1
 */
class FARSA_PLUGIN_API CartPoleExperiment : public farsa::EvoRobotExperiment
{
	Q_OBJECT
	FARSA_REGISTER_CLASS(EvoRobotExperiment)

public:
	/**
	 * \brief Constructor
	 */
    CartPoleExperiment();

	/**
	 * \brief Destructor
	 */
    ~CartPoleExperiment();

	/**
	 * \brief Configures the object using a ConfigurationParameters object
	 *
	 * \param params the configuration parameters object with parameters to
	 *               use
	 * \param prefix the prefix to use to access the object configuration
	 *               parameters. This is guaranteed to end with the
	 *               separator character when called by the factory, so you
	 *               don't need to add one
	 */
	virtual void configure(farsa::ConfigurationParameters& params, QString prefix);

	/**
	 * \brief Saves the actual status of parameters into the
	 *        ConfigurationParameters object passed
	 *
	 * \param params the configuration parameters object on which to save
	 *               the actual parameters
	 * \param prefix the prefix to use to access the object configuration
	 *               parameters
	 */
	virtual void save(farsa::ConfigurationParameters& params, QString prefix);

	/**
	 * \brief Add to Factory::typeDescriptions() the descriptions of all
	 *        parameters and subgroups
	 *
	 * It's mandatory in all subclasses where configure and save methods
	 * have been re-implemented for dealing with new parameters and
	 * subgroups to also implement the describe method
	 * \param type is the name of the type regarding the description. The
	 *             type is used when a subclass reuse the description of its
	 *             parent calling the parent describe method passing the
	 *             type of the subclass. In this way, the result of the
	 *             method describe of the parent will be the addition of the
	 *             description of the parameters of the parent class into
	 *             the type of the subclass
	 */
	static void describe( QString type );

	/**
	 * \brief This function is called after all linked objects have been
	 *        configured
	 *
	 * See the description of the ConfigurationParameters class for more
	 * information. This method creates the arena and fills it with objects
	 */
	virtual void postConfigureInitialization();

	/**
	 * \brief Called at the begin of each generation
	 *
	 * \param generation the generation about to start
	 */
	virtual void initGeneration(int generation);

	/**
	 * \brief Called before the evaluation of a new individual
	 *
	 * \param individual the id of the individual about to be tested
	 */
	virtual void initIndividual(int individual);

	/**
	 * \brief Called at the beginning of each trial
	 *
	 * \param trial the trial about to start
	 */
	virtual void initTrial(int trial);

	/**
	 * \brief Called at the beginning of each step (before world advance)
	 *
	 * \param step the step about to start
	 */
	virtual void initStep(int step);

	/**
	 * \brief Celled after all sensors have been updated but before network
	 *        spreading
	 *
	 * This is useful, for example, to overwrite the inputs of the neural
	 * network (i.e.: to silence some neurons during the experiment without
	 * modifing sensors classes)
	 */
	virtual void afterSensorsUpdate();

	/**
	 * \brief Called just before updating motors and after updating the
	 *        neural network
	 *
	 * This is useful, for example, to overwrite the outputs of the neural
	 * network
	 */
	virtual void beforeMotorsUpdate();

	/**
	 * \brief Called just before the world advances, after the update of
	 *        motors
	 *
	 * This is useful, for example, to manually actuate motors overriding
	 * the robot controller commands
	 */
	virtual void beforeWorldAdvance();

	/**
	 * \brief Called at the end of each step
	 *
	 * \param step the step about to end
	 */
	virtual void endStep(int step);

	/**
	 * \brief Called at the end of each trial
	 *
	 * \param trial the trial about to end
	 */
	virtual void endTrial(int trial);

	/**
	 * \brief Called at the end of an individual life
	 *
	 * \param individual the individual that has been just tested
	 */
	virtual void endIndividual(int individual);

	/**
	 * \brief Called at the end of each generation
	 *
	 * \param generation the generation about to end
	 */
	virtual void endGeneration(int generation);

	//! Get a pointer to the random generator
	farsa::RandomGenerator* getRng();

private:

	//! Generalization test
	double testCartPole(farsa::Evonet* evonet);

	//! Set the seed of the random generator
	void setRngSeed(const int seed);

	//! The class that implements the Cart-Pole
	class CartPole {
	public:
		CartPole();
		~CartPole();
		void initializeCartPole(bool randomize,bool velocity,farsa::EvoRobotExperiment* exp);
		virtual void simplifyTask();
		virtual void nextTask();
		virtual double evalNet(farsa::Evonet* evonet,int thresh);
		void setDelay(const int delay);
		void setStateRanges(QVector<double> ranges);
		void setJigglingScale(const double scale);
		void resetAll();
		void useFixedPositions(const bool flag);
		int getCompleted();
		void saveTrajectory(const int individual);
		void setSecondPole(const double len, const double mass);
		double maxFitness;
		bool MARKOV;
		
		bool last_hundred;
		bool nmarkov_long;  //Flag that we are looking at the champ
		bool generalization_test;  //Flag we are testing champ's generalization
		bool testCart;
		QString stampOuput;
		
		double state[6];
		
		double jigglestep[1000];
	
	protected:
		virtual void init(int type);
	
	private:
	
		void performAction(double output, int stepnum);
		void step(double action, double *state, double *derivs);
		void rk4(double f, double y[], double dydx[], double yout[]);
		bool outsideBounds();
		
		const static int NUM_INPUTS=7;
		//const static double MUP = 0.000002;
		//const static double MUC = 0.0005;
		//const static double GRAVITY= -9.8;
		//const static double MASSCART= 1.0;
		//const static double MASSPOLE_1= 0.1;
		
		//const static double LENGTH_1= 0.5;		  /* actually half the pole's length */
		
		//const static double FORCE_MAG= 10.0;
		//const static double TAU= 0.01;		  //seconds between state updates
		
		//const static double one_degree= 0.0174532;	/* 2pi/360 */
		//const static double six_degrees= 0.1047192;
		//const static double twelve_degrees= 0.2094384;
		//const static double fifteen_degrees= 0.2617993;
		//const static double thirty_six_degrees= 0.628329;
		//const static double fifty_degrees= 0.87266;

		const double MUP;
		const double MUC;
		const double GRAVITY;
		const double MASSCART;
		const double MASSPOLE_1;
		
		const double LENGTH_1;		  /* actually half the pole's length */
		
		const double FORCE_MAG;
		const double TAU;		  //seconds between state updates
		
		const double one_degree;	/* 2pi/360 */
		const double six_degrees;
		const double twelve_degrees;
		const double fifteen_degrees;
		const double thirty_six_degrees;
		const double fifty_degrees;
		
		double LENGTH_2;
		double MASSPOLE_2;
		double MIN_INC;
		double POLE_INC;
		double MASS_INC;
		
		//Queues used for Gruau's fitness which damps oscillations
		int balanced_sum;
		double cartpos_sum;
		double cartv_sum;
		double polepos_sum;
		double polev_sum;

		QVector<double> stateRanges;
		int delay;
		farsa::EvoRobotExperiment* experiment;

		bool stdFit;
		int completed;
		int jigglingScale;
		bool fixed;
		bool saveTraj;
		int localInd;
	};
	
	// The cart pole used for evaluating individuals
	CartPole m_cartPole;
	// Delayed task
	bool m_delayTask;
	int m_delay;
	// Flags whether or not generalization test must be run
	bool m_test;
	int m_testTrials;
	// Random number generator
	farsa::RandomGenerator m_rng;
	// Local seed
	int m_seed;
	//! State range values
	double m_rangeState0;
	double m_rangeState1;
	double m_rangeState2;
	double m_rangeState3;
	bool m_specifySecondPoleStates;
	double m_rangeState4;
	double m_rangeState5;
	//! Bonus for completing trials
	bool m_bonus;
	//! Scale of the exponential component (jiggling)
	double m_jigglingScale;
	//! Flags whether or not the fitness must be averaged over the number of trials
	bool m_normalizeFitness;
	//! Flags whether or not the evolution must be run with 8 fixed positions
	bool m_fixedPositions;
	//! Completed trials
	int m_completedTrials;
	//! Save trajectory
	bool m_saveTrajectory;
	//! Second pole's length
	double m_secondPoleLength;
	//! Second pole's mass
	double m_secondPoleMass;
	//! Current generation
	int m_currGen;
};

#endif
