/********************************************************************************
 *  FARSA - Total99                                                             *
 *  Copyright (C) 2012-2013 Gianluca Massera <emmegian@yahoo.it>                *
 *                                                                              *
 *  This program is free software; you can redistribute it and/or modify        *
 *  it under the terms of the GNU General Public License as published by        *
 *  the Free Software Foundation; either version 2 of the License, or           *
 *  (at your option) any later version.                                         *
 *                                                                              *
 *  This program is distributed in the hope that it will be useful,             *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *  GNU General Public License for more details.                                *
 *                                                                              *
 *  You should have received a copy of the GNU General Public License           *
 *  along with this program; if not, write to the Free Software                 *
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA  *
 ********************************************************************************/

#ifndef MYEVOALGO_H
#define MYEVOALGO_H

#include "farsaplugin.h"

#include "evoga.h"
#include "logger.h"
#include "evodataviewer.h"
#include "randomgenerator.h"
#include "evorobotexperiment.h"
#include "evonet.h"
#include <parametersettable.h>
#include <configurationhelper.h>
#include <configurationparameters.h>

#include <vector>
#include <fstream>
#include <iostream>
#include <deque>
#include <iomanip>
#include <sstream>
#include <string>

#include <QFile>
#include <QTime>
#include <QVector>
#include <QtAlgorithms>
#include <QThreadPool>
#include <QtConcurrentMap>

/**
 * \brief My evolutionary algorithm. It inherits from Evoga
 *        and runs a modified version of the steadyState
 *        algorithm (with floating-point genes).
 */
class FARSA_PLUGIN_API MyEvoAlgo : public farsa::Evoga
{
	Q_OBJECT
	FARSA_REGISTER_CLASS(Evoga)

	struct Gene 
	{
		int input;
		int output;
		float weight;
		bool enabled;
	};

	struct NeuralNetworkInfo
	{
		int numNeurons;
		int numInputs;
		int numOutputs;
		bool inputOutputConnections;
		bool recurrentHiddens;
		bool recurrentOutputs;
		float weightRange;
	};

public:
     /**
	 * \brief Constructor
	 */
	 MyEvoAlgo();


     /**
	 * \brief Destructor
	 */     
    ~MyEvoAlgo();
	
	/**
	 * \brief Configures the object using a ConfigurationParameters object
	 *
	 * \param params the configuration parameters object with parameters to
	 *               use
	 * \param prefix the prefix to use to access the object configuration
	 *               parameters. This is guaranteed to end with the
	 *               separator character when called by the factory, so you
	 *               don't need to add one
	 */
	virtual void configure(farsa::ConfigurationParameters& params, QString prefix);

	virtual void save(farsa::ConfigurationParameters& params, QString prefix)
	{
		// Calling parent function
		farsa::Evoga::save(params, prefix);
		farsa::Logger::warning("NOT IMPLEMENTED (MyEvoAlgo::save)");
		abort();
	}

	//! Describe
	static void describe(QString type);
	
	void postConfigureInitialization();
	/**
	 * \brief Entry point of MyEvoAlgo class
	 */
	void xnesAlgo();

	//! Post evaluation test (Gruau's generalization score)
	void postEvaluationTest();
	/**
	 * \brief Generalization test of the evolved individuals
	 */
	void generalizationTest();
	/**
	 * \brief Detailed test of the evolved individuals
	 */
	void detailedTest();
	/**
	 * \brief Behavioural analysis
	 */
	void analyzeBehavior();
	/**
	 * \brief Analyze weights
	 */
	void analyzeWeights();
	void computeIndividualFitness();
	/**
	 * \brief It invokes either evolve() or test() method independently of
	 *        the chosen type. This depends on the <test> flag.
	 */
	virtual void evolveAllReplicas();

	void saveBestInd(int bestind);

	void saveIndividual(const float* ind);

	void saveIndividualFitness(const double fit);

	void saveEvolved(const float* ind);

	void saveagenotype(FILE *fp, int ind);

	void saveallg();

	void loadgenotype(FILE *fp, int ind);

	int loadallg(int gen, const char *filew);

	void computeFStat();

	void saveFStat();

	/**
     * \brief Loads genotypes from the given file
     *
     * \param filename the name of the file to load
     * \return the number of individuals that have been loaded, 0 in case of
     *         errors
     */
    unsigned int loadGenotypes(QString filename);

    /**
     * \brief Returns number of loaded genotypes
     *
     * \return the number of individuals that have been loaded
     */
    unsigned int numLoadedGenotypes() const;

	int getLoadedIndividualLength(const int id);

	/**
     * \brief Returns the array of genes for the given individual
     *
     * \param ind the individual for whom genes are requested
     * \return the array of genes for the given individual
     */
    QVector<float> getFloatVectorGenesForIndividual(unsigned int id);

	/*! Initialize the random number generator
     *
     *  \param s The seed use dto initialize the random number generator
     */
    void setSeed(int s);

private:
	/**
	 * \brief Returns the random number generator
	 *
	 * \return The random number generator
	 */
	farsa::RandomGenerator* getRandomGenerator();
	/**
	 * \brief Round a double value to the nearest integer
	 *
	 * \param value The value
	 * \return The nearest integer
	 */
	int round(double value);

	//! Population
	QVector<float *> m_pop;
	//! Local random number generator
	farsa::RandomGenerator m_rng;
	//! Mutation range
	double m_mutationRange;
	//! Flags whether or not a test must be performed
	bool m_test;
	//! Flags whether or not balancing test must be run
	bool m_generalizationTest;
	//! Flags whether or not detailed test must be run
	bool m_detailedTest;
	//! Flags whether or not the behavioural analysis must be run
	bool m_behaviourAnalysis;
	//! Compute individual's fitness
	bool m_calcIndFit;
	//! How many generations the best individual is stored
	int m_saveStatGenCnt;
	//! Standard deviation of the Gaussian distribution
	double m_gaussianStdDev;
	//! Step rate
	double m_stepRate;
    //! The probability to accept offsprings that received maladaptive variations
    double m_stochasticity;
	//! Score flag
	bool m_scoreFlag;
	//! Flags whether or not a custom number of samples must be used
	bool m_defineOffspring;
	//! Custom number of samples (i.e., offspring)
	int m_numOffspring;
	//! Analyze weights
	bool m_weightAnalysis;
	//! How many generations a statistical test on generalization capability is run
	int m_statTestGen;
	//! Number of evaluations
	int m_numEvals;
	//! Number of statistical evaluations
	int m_numStatEvals;
	//! Number of times the old individual is placed back
	int m_numReuses;
};

#endif
