/********************************************************************************
 *  FARSA - Total99                                                             *
 *  Copyright (C) 2005-2011 Gianluca Massera <emmegian@yahoo.it>                *
 *                                                                              *
 *  This program is free software; you can redistribute it and/or modify        *
 *  it under the terms of the GNU General Public License as published by        *
 *  the Free Software Foundation; either version 2 of the License, or           *
 *  (at your option) any later version.                                         *
 *                                                                              *
 *  This program is distributed in the hope that it will be useful,             *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *  GNU General Public License for more details.                                *
 *                                                                              *
 *  You should have received a copy of the GNU General Public License           *
 *  along with this program; if not, write to the Free Software                 *
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA  *
 ********************************************************************************/

#include "cartpoleexperiment.h"
#include "evoga.h"
#include "utilitiesexceptions.h"
#include "randomgenerator.h"
#include "robots.h"
#include "world.h"
#include "phybox.h"
#include "wheeledexperimenthelper.h"
#include "logger.h"
#include "configurationhelper.h"

// This is needed because the isnan and isinf functions are not present windows
#ifdef WIN32
	#include <float.h>
	#define isnan(x) _isnan(x)
	#define isinf(x) (!_finite(x))
#else
	#define isnan(x) std::isnan(x)
	#define isinf(x) std::isinf(x)
#endif

CartPoleExperiment::CartPoleExperiment() :
	farsa::EvoRobotExperiment(),
	m_cartPole(),
	m_delay(0),
	m_test(false),
	m_testTrials(0),
	m_rng(1),
	m_seed(1),
	m_rangeState0(0.0),
	m_rangeState1(0.0),
	m_rangeState2(0.0),
	m_rangeState3(0.0),
	m_specifySecondPoleStates(false),
	m_rangeState4(0.0),
	m_rangeState5(0.0),
	m_bonus(false),
	m_jigglingScale(1.0),
	m_normalizeFitness(false),
	m_fixedPositions(false),
	m_completedTrials(0),
	m_saveTrajectory(false),
	m_secondPoleLength(0.05),
	m_secondPoleMass(0.01),
	m_currGen(-1)
{
	// No need to add resources, they are added by EvoRobotExperiment
}

CartPoleExperiment::~CartPoleExperiment()
{
}

void CartPoleExperiment::configure(farsa::ConfigurationParameters& params, QString prefix)
{
	// Calling parent function
	farsa::EvoRobotExperiment::configure(params, prefix);
	m_delay = farsa::ConfigurationHelper::getInt(params, prefix + "delay", m_delay);
	m_test = farsa::ConfigurationHelper::getBool(params, prefix + "test", m_test);
	if (m_test)
	{
		m_testTrials = farsa::ConfigurationHelper::getInt(params, prefix + "testTrials", m_testTrials);
	}
	m_rangeState0 = farsa::ConfigurationHelper::getDouble(params, prefix + "rangeState0", m_rangeState0);
	m_rangeState1 = farsa::ConfigurationHelper::getDouble(params, prefix + "rangeState1", m_rangeState1);
	m_rangeState2 = farsa::ConfigurationHelper::getDouble(params, prefix + "rangeState2", m_rangeState2);
	m_rangeState3 = farsa::ConfigurationHelper::getDouble(params, prefix + "rangeState3", m_rangeState3);
	m_specifySecondPoleStates = farsa::ConfigurationHelper::getBool(params, prefix + "specifySecondPoleStates", m_specifySecondPoleStates);
	if (m_specifySecondPoleStates)
	{
		m_rangeState4 = farsa::ConfigurationHelper::getDouble(params, prefix + "rangeState4", m_rangeState4);
		m_rangeState5 = farsa::ConfigurationHelper::getDouble(params, prefix + "rangeState5", m_rangeState5);
	}
	m_bonus = farsa::ConfigurationHelper::getBool(params, prefix + "bonus", m_bonus);
	m_jigglingScale = farsa::ConfigurationHelper::getDouble(params, prefix + "jigglingScale", m_jigglingScale);
	if (m_jigglingScale == 0.0)
	{
		farsa::Logger::warning(QString("The scale for jiggling could not be 0. We set the default value!"));
		m_jigglingScale = 1.0;
	}
	m_fixedPositions = farsa::ConfigurationHelper::getBool(params, prefix + "fixedPositions", m_fixedPositions);
	m_secondPoleLength = farsa::ConfigurationHelper::getDouble(params, prefix + "secondPoleLength", m_secondPoleLength);
	m_secondPoleMass = farsa::ConfigurationHelper::getDouble(params, prefix + "secondPoleMass", m_secondPoleMass);
}

void CartPoleExperiment::save(farsa::ConfigurationParameters& params, QString prefix)
{
	// Calling parent function
	farsa::EvoRobotExperiment::save(params, prefix);

    farsa::Logger::error("NOT IMPLEMENTED (CartPoleExperiment::save)");
	abort();
}

void CartPoleExperiment::describe(QString type)
{
	// Calling parent function
	farsa::EvoRobotExperiment::describe(type);

	Descriptor d = addTypeDescription(type, "The cart-pole experiment");
	d.describeInt("delay").limits(0,INT_MAX).def(0).help("The amount of delay");
	d.describeBool("test").def(false).help("Flags whether or not generalization test must be run");
	d.describeInt("testTrials").limits(0,INT_MAX).def(0).help("The number of trials performed during the generalization test");
	d.describeReal("rangeState0").def(1.0).help("Coefficient of state 0");
	d.describeReal("rangeState1").def(1.0).help("Coefficient of state 1");
	d.describeReal("rangeState2").def(1.0).help("Coefficient of state 2");
	d.describeReal("rangeState3").def(1.0).help("Coefficient of state 3");
}

void CartPoleExperiment::postConfigureInitialization()
{
	// Calling parent function
	EvoRobotExperiment::postConfigureInitialization();

	m_cartPole.nmarkov_long = false;
    m_cartPole.generalization_test = false;
    m_cartPole.testCart = false;
    m_cartPole.initializeCartPole(true,false,this);

	// Set the seed
	farsa::Evoga* ga = getEvoga();
	const int m_seed = ga->getCurrentSeed();
	setRngSeed(m_seed);
	// Set variables of the cart-pole
	QVector<double> ranges(6);
	ranges[0] = m_rangeState0;
	ranges[1] = m_rangeState1;
	ranges[2] = m_rangeState2;
	ranges[3] = m_rangeState3;
	ranges[4] = m_rangeState4;
	ranges[5] = m_rangeState5;
	m_cartPole.setStateRanges(ranges);
	if (m_delay > 0)
	{
		// Delay
		m_cartPole.setDelay(m_delay);
	}
	m_cartPole.setJigglingScale(m_jigglingScale);
	m_cartPole.useFixedPositions(m_fixedPositions);
	m_cartPole.setSecondPole(m_secondPoleLength, m_secondPoleMass);
}

void CartPoleExperiment::initGeneration(int generation)
{
	m_currGen = generation;
}

void CartPoleExperiment::initIndividual(int individual)
{
	if (individual <= -9999)
	{
		// Set the number of trials to 50
		setNTrials(50);
		// Save trajectory
		m_saveTrajectory = true;
		m_cartPole.saveTrajectory(-9999 - individual);
	}
	if ((individual > -9999) && (individual < 0))
	{
		// Set the number of trials to 1000
		setNTrials(1000);
		// Set the flag to true
		m_normalizeFitness = true;
		m_completedTrials = 0;
	}
	if (individual >= 9999)
	{
		// Save trajectory
		m_saveTrajectory = true;
		m_cartPole.saveTrajectory(individual - 9999);
		m_cartPole.useFixedPositions(true);
	}
}

void CartPoleExperiment::initTrial(int /*trial*/)
{
	trialFitnessValue = 0.0;
	m_cartPole.resetAll();
}

void CartPoleExperiment::initStep(int /*step*/)
{
}

void CartPoleExperiment::afterSensorsUpdate()
{
}

void CartPoleExperiment::beforeMotorsUpdate()
{
}

void CartPoleExperiment::beforeWorldAdvance()
{
}

void CartPoleExperiment::endStep(int /*step*/)
{
}

void CartPoleExperiment::endTrial(int trial)
{
	farsa::ResourcesLocker locker(this);
	farsa::Evonet* evonet = getResource<farsa::Evonet>("evonet");
	evonet->resetNet();
	if (m_test)
	{
		// Perform generalization test
		trialFitnessValue = testCartPole(evonet);
		totalFitnessValue = trialFitnessValue;
	}
	else
	{
		// Evolution
		trialFitnessValue = m_cartPole.evalNet(evonet, trial);
		//totalFitnessValue += trialFitnessValue / farsa::real(getNSteps());
		totalFitnessValue += trialFitnessValue; // We do not have steps
		if (m_normalizeFitness)
		{
			m_completedTrials += m_cartPole.getCompleted();
		}
	}
}

void CartPoleExperiment::endIndividual(int /*individual*/)
{
	if (m_normalizeFitness)
	{
		totalFitnessValue = totalFitnessValue / farsa::real(getNTrials());
		// Save the number of trials
		char filename[64];
		sprintf(filename, "genCompletedTrials.txt");
		FILE* fp = fopen(filename, "a");
		if (fp != NULL)
		{
			fprintf(fp, "%d\n", m_completedTrials);
			fclose(fp);
		}
	}
	if (m_saveTrajectory)
	{
		totalFitnessValue = totalFitnessValue / farsa::real(getNTrials());
	}
}

void CartPoleExperiment::endGeneration(int /*generation*/)
{
}

double CartPoleExperiment::testCartPole(farsa::Evonet* evonet)
{
	double score = 0.0;
	double fit;
	double statevals[5] = {0.05, 0.25, 0.5, 0.75, 0.95};
    int s0c, s1c, s2c, s3c;
    for (s0c = 0; s0c <= 4; ++s0c)
	{
		for (s1c = 0; s1c <= 4; ++s1c)
		{
			for (s2c = 0; s2c <= 4; ++s2c)
			{
				for (s3c = 0; s3c <= 4; ++s3c) 
				{
					m_cartPole.state[0] = statevals[s0c] * 4.32 - 2.16;
					m_cartPole.state[1] = statevals[s1c] * 2.70 - 1.35;
					m_cartPole.state[2] = statevals[s2c] * 0.12566304 - 0.06283152;
					/* 0.06283152 =  3.6 degrees */
					m_cartPole.state[3] = statevals[s3c] * 0.30019504 - 0.15009752;
					/* 0.15009752 =  8.6 degrees */
					m_cartPole.state[4]=0.0;
					m_cartPole.state[5]=0.0;
					m_cartPole.generalization_test=true;
					// Compute the fitness network
					fit = m_cartPole.evalNet(evonet,1);
					if (fit >= 999) 
					{
						score++;
					}
				}
			}
		}
	}
	return score;
}

void CartPoleExperiment::setRngSeed(const int seed)
{
	m_rng.setSeed(seed);
}

farsa::RandomGenerator* CartPoleExperiment::getRng()
{
	return &m_rng;
}

// The following source code (lines 316-749) has been taken from http://nn.cs.utexas.edu/keyword?neat-c (Stanley and Miikkulainen, 2002)
CartPoleExperiment::CartPole::CartPole()
	: MUP(0.000002)
	, MUC(0.0005)
	, GRAVITY(-9.8)
	, MASSCART(1.0)
	, MASSPOLE_1(0.1)
	, LENGTH_1(0.5)
	, FORCE_MAG(10.0)
	, TAU(0.01)
	, one_degree(0.0174532)
	, six_degrees(0.1047192)
	, twelve_degrees(0.2094384)
	, fifteen_degrees(0.2617993)
	, thirty_six_degrees(0.628329)
	, fifty_degrees(0.87266)
{
}

CartPoleExperiment::CartPole::~CartPole()
{
}

void CartPoleExperiment::CartPole::initializeCartPole(bool randomize, bool velocity, farsa::EvoRobotExperiment* exp)
{
	maxFitness = 100000;
	delay = 0;
	experiment = exp;
	stdFit = true;
	jigglingScale = 1.0;
	completed = 0;
	fixed = false;
	saveTraj = false;
	localInd = -1;
	
	MARKOV=velocity;
	stampOuput = "";
	
	MIN_INC = 0.001;
	POLE_INC = 0.05;
	MASS_INC = 0.01;
	
	LENGTH_2 = 0.05;
	MASSPOLE_2 = 0.01;
}

double CartPoleExperiment::CartPole::evalNet(farsa::Evonet *evonet, int thresh)
{
	int steps=0;
	double input[NUM_INPUTS];
	double output;
	QVector<QVector<double> > stepMatrix;
	
	int nmarkovmax;
	
	double nmarkov_fitness;
	
	double jiggletotal; //total jiggle in last 100
	int count;  //step counter
	
	// Number of steps
	nmarkovmax = 1000;
	
	// Initialize trial
	init(thresh);
	
	//NON MARKOV CASE
	while (steps++ < nmarkovmax) 
	{
		// Set the state
		QVector<double> tmpState;
		tmpState.append(state[0] / 4.8);
		tmpState.append(state[2]  / 0.52);
		tmpState.append(state[4] / 0.52);
		tmpState.append(.5);
	
		stepMatrix.append(tmpState);
	
		if(steps - 1 < delay)
		{
			input[0] = stepMatrix[steps-1][0];
			input[1] = stepMatrix[steps-1][1];
			input[2] = stepMatrix[steps-1][2];
			input[3] = stepMatrix[steps-1][3];
		}
		else
		{
			input[0] = stepMatrix[steps-1-delay][0];
			input[1] = stepMatrix[steps-1-delay][1];
			input[2] = stepMatrix[steps-1-delay][2];
			input[3] = stepMatrix[steps-1-delay][3];
	
		}
	
		// Set the network input
		for(int j = 0; j < 4 ; j++)
		{
			evonet->setInput(j,input[j]);
		}
	
		// Activate the net
		evonet->updateNet();
		// Get the network output
		output=evonet->getOutput(0);
		// Perform action (i.e., the new state)
		performAction(output,steps);

		if (saveTraj)
		{
			char filename[64];
			sprintf(filename, "trajectoryS%dTrial%d.txt", localInd, thresh);
			FILE* fp = fopen(filename, "a");
			if (fp != NULL)
			{
				fprintf(fp, "%lf\n", state[0]);
				fclose(fp);
			}
		}
	
		if (outsideBounds())	// if failure
			break;			// stop it now
	}
	
	//If we are generalizing we just need to balance it a while
	if (generalization_test)
	{
		return (double) balanced_sum;
	}
	
	//Sum last 100
	if ((steps>100) && (!nmarkov_long)) {
		jiggletotal=0;
		//Adjust for array bounds and count
		for (count=steps-99-2;count<=steps-2;count++)
		{
			jiggletotal+=jigglestep[count];
		}
	}
	
	if (balanced_sum>100)
	{
		nmarkov_fitness = ((double) balanced_sum) / 1000.0;
		if (balanced_sum == 1000)
		{
			completed++;
		}
	}
	else
	{
		nmarkov_fitness=((((double) balanced_sum)/1000.0));
		if (balanced_sum == 1000)
		{
			completed++;
		}
	}
	return nmarkov_fitness;
}

void CartPoleExperiment::CartPole::init(int type)
{

  if (!MARKOV) {
    //Clear all fitness records
    cartpos_sum=0.0;
    cartv_sum=0.0;
    polepos_sum=0.0;
    polev_sum=0.0;
  }

  balanced_sum=0; //Always count # balanced

  last_hundred=false;

  if (!generalization_test) {
	  
	  if (!fixed)
	  {
		  CartPoleExperiment* localExp = dynamic_cast<CartPoleExperiment*>(experiment);

		  for (int i = 0; i < stateRanges.size(); i++)
		  {
			  state[i] = (localExp->getRng())->getDouble(-stateRanges[i],stateRanges[i]);
		  }
	  }
	  else
	  {
		if (type == 0)
		{
			state[0] = -stateRanges[0];
			state[1] = state[2] = state[3] = state[4] = state[5] = 0;
		}
		if (type == 1)
		{
			state[0] = stateRanges[0];
			state[1] = state[2] = state[3] = state[4] = state[5] = 0;
		}
		if (type == 2)
		{
			state[1] = -stateRanges[1];
			state[0] = state[2] = state[3] = state[4] = state[5] = 0;
		}
		if (type == 3)
		{
			state[1] = stateRanges[1];
			state[0] = state[2] = state[3] = state[4] = state[5] = 0;
		}
		if (type == 4)
		{
			state[2] = -stateRanges[2];
			state[0] = state[1] = state[3] = state[4] = state[5] = 0;
		}
		if (type == 5)
		{
			state[2] = stateRanges[2];
			state[0] = state[1] = state[3] = state[4] = state[5] = 0;
		}
		if (type == 6)
		{
			state[3] = -stateRanges[3];
			state[0] = state[1] = state[2] = state[4] = state[5] = 0;
		}
		if (type == 7)
		{
			state[3] = stateRanges[3];
			state[0] = state[1] = state[2] = state[4] = state[5] = 0;
		}
	  }
  }
  else {
    state[4] = state[5] = 0;
  }
}

void CartPoleExperiment::CartPole::performAction(double output, int stepnum)
{

  int i;
  double  dydx[6];

  const bool RK4=true; //Set to Runge-Kutta 4th order integration method
  const double EULER_TAU= TAU/4;

    if(testCart)
        stampOuput += QString("%1 %2 %3 \n").arg(state[0]).arg(state[2]).arg(state[4]);

  /*--- Apply action to the simulated cart-pole ---*/

  if(RK4){
    for(i=0;i<2;++i){
      dydx[0] = state[1];
      dydx[2] = state[3];
      dydx[4] = state[5];
      step(output,state,dydx);
      rk4(output,state,dydx,state);
    }
  }
  else{
    for(i=0;i<8;++i){
      step(output,state,dydx);
      state[0] += EULER_TAU * dydx[0];
      state[1] += EULER_TAU * dydx[1];
      state[2] += EULER_TAU * dydx[2];
      state[3] += EULER_TAU * dydx[3];
      state[4] += EULER_TAU * dydx[4];
      state[5] += EULER_TAU * dydx[5];
    }
  }

  //Record this state
  cartpos_sum+=fabs(state[0]);
  cartv_sum+=fabs(state[1]);
  polepos_sum+=fabs(state[2]);
  polev_sum+=fabs(state[3]);
  if (stepnum<=1000)
    jigglestep[stepnum-1]=fabs(state[0])+fabs(state[1])+fabs(state[2])+fabs(state[3]);

  if (false) {
    if (!(outsideBounds())) {
      if (balanced_sum<1000) {
    //cout<<".";
    ++balanced_sum;
      }
    }
    else {
      if (balanced_sum==1000)
    balanced_sum=1000;
      else balanced_sum=0;
    }
  }
  else if (!(outsideBounds()))
    ++balanced_sum;

}

void CartPoleExperiment::CartPole::step(double action, double *st, double *derivs)
{
    double force,costheta_1,costheta_2,sintheta_1,sintheta_2,
          gsintheta_1,gsintheta_2,temp_1,temp_2,ml_1,ml_2,fi_1,fi_2,mi_1,mi_2;

    force =  (action - 0.5) * FORCE_MAG * 2;
    costheta_1 = cos(st[2]);
    sintheta_1 = sin(st[2]);
    gsintheta_1 = GRAVITY * sintheta_1;
    costheta_2 = cos(st[4]);
    sintheta_2 = sin(st[4]);
    gsintheta_2 = GRAVITY * sintheta_2;

    ml_1 = LENGTH_1 * MASSPOLE_1;
    ml_2 = LENGTH_2 * MASSPOLE_2;
    temp_1 = MUP * st[3] / ml_1;
    temp_2 = MUP * st[5] / ml_2;
    fi_1 = (ml_1 * st[3] * st[3] * sintheta_1) +
           (0.75 * MASSPOLE_1 * costheta_1 * (temp_1 + gsintheta_1));
    fi_2 = (ml_2 * st[5] * st[5] * sintheta_2) +
           (0.75 * MASSPOLE_2 * costheta_2 * (temp_2 + gsintheta_2));
    mi_1 = MASSPOLE_1 * (1 - (0.75 * costheta_1 * costheta_1));
    mi_2 = MASSPOLE_2 * (1 - (0.75 * costheta_2 * costheta_2));

    derivs[1] = (force + fi_1 + fi_2)
                 / (mi_1 + mi_2 + MASSCART);

    derivs[3] = -0.75 * (derivs[1] * costheta_1 + gsintheta_1 + temp_1)
                 / LENGTH_1;
    derivs[5] = -0.75 * (derivs[1] * costheta_2 + gsintheta_2 + temp_2)
                  / LENGTH_2;

}

void CartPoleExperiment::CartPole::rk4(double f, double y[], double dydx[], double yout[])
{

    int i;

    double hh,h6,dym[6],dyt[6],yt[6];


    hh=TAU*0.5;
    h6=TAU/6.0;
    for (i=0;i<=5;i++) yt[i]=y[i]+hh*dydx[i];
    step(f,yt,dyt);
    dyt[0] = yt[1];
    dyt[2] = yt[3];
    dyt[4] = yt[5];
    for (i=0;i<=5;i++) yt[i]=y[i]+hh*dyt[i];
    step(f,yt,dym);
    dym[0] = yt[1];
    dym[2] = yt[3];
    dym[4] = yt[5];
    for (i=0;i<=5;i++) {
        yt[i]=y[i]+TAU*dym[i];
        dym[i] += dyt[i];
    }
    step(f,yt,dyt);
    dyt[0] = yt[1];
    dyt[2] = yt[3];
    dyt[4] = yt[5];
    for (i=0;i<=5;i++)
        yout[i]=y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);
}

bool CartPoleExperiment::CartPole::outsideBounds()
{
  const double failureAngle = thirty_six_degrees;

  return
    state[0] < -2.4              ||
    state[0] > 2.4               ||
    state[2] < -failureAngle     ||
    state[2] > failureAngle      ||
    state[4] < -failureAngle     ||
    state[4] > failureAngle;
}

void CartPoleExperiment::CartPole::nextTask()
{

   LENGTH_2 += POLE_INC;   /* LENGTH_2 * INCREASE;   */
   MASSPOLE_2 += MASS_INC; /* MASSPOLE_2 * INCREASE; */
}

void CartPoleExperiment::CartPole::simplifyTask()
{
  if(POLE_INC > MIN_INC) {
    POLE_INC = POLE_INC/2;
    MASS_INC = MASS_INC/2;
    LENGTH_2 -= POLE_INC;
    MASSPOLE_2 -= MASS_INC;
  }
}

void CartPoleExperiment::CartPole::setDelay(const int delay)
{
	this->delay = delay;
}

void CartPoleExperiment::CartPole::setStateRanges(QVector<double> ranges)
{
	stateRanges.resize(ranges.size());
	for (int i = 0; i < stateRanges.size(); i++)
	{
		stateRanges[i] = ranges[i];
	}
}

void CartPoleExperiment::CartPole::setJigglingScale(const double scale)
{
	this->jigglingScale = scale;
}

void CartPoleExperiment::CartPole::resetAll()
{
	completed = 0;
}

int CartPoleExperiment::CartPole::getCompleted()
{
	return completed;
}

void CartPoleExperiment::CartPole::useFixedPositions(const bool flag)
{
	this->fixed = flag;
}

void CartPoleExperiment::CartPole::saveTrajectory(const int individual)
{
	this->saveTraj = true;
	this->localInd = individual;
}

void CartPoleExperiment::CartPole::setSecondPole(const double len, const double mass)
{
	this->LENGTH_2 = len;
	this->MASSPOLE_2 = mass;
}
