#include "myevoalgo.h"
#include "evonet.h"
#include <cmath>
#include <QTextStream>
#include <QFile>

#include <Eigen/Core>
#include <Eigen/Dense>
#include <unsupported/Eigen/MatrixFunctions>

using namespace std;

//------------------------------------- MYEVOALGO CLASS ------------------------------------------------------
//---------------------------------- Inherited from EVOGA ----------------------------------------------------

MyEvoAlgo::MyEvoAlgo()
	: farsa::Evoga()
	, m_pop()
	, m_rng(1)
    , m_mutationRange(0.5)
	, m_stepRate(1.0)
	, m_test(false)
	, m_generalizationTest(false)
	, m_detailedTest(false)
	, m_behaviourAnalysis(false)
	, m_calcIndFit(false)
	, m_saveStatGenCnt(0)
	, m_gaussianStdDev(0.1)
	, m_scoreFlag(false)
	, m_defineOffspring(false)
	, m_numOffspring(0)
	, m_weightAnalysis(false)
	, m_statTestGen(0)
	, m_numEvals(0)
{
}

MyEvoAlgo::~MyEvoAlgo()
{
}

void MyEvoAlgo::configure(farsa::ConfigurationParameters& params, QString prefix)
{
	// Calling parent function
	farsa::Evoga::configure(params, prefix);
    m_mutationRange = farsa::ConfigurationHelper::getDouble(params, prefix + "mutationRange", m_mutationRange);
	m_stepRate = farsa::ConfigurationHelper::getDouble(params, prefix + "stepRate", m_stepRate);
	if (m_stepRate == 0.0)
	{
		// Set to default
		m_stepRate = 1.0;
	}
	m_test = farsa::ConfigurationHelper::getBool(params, prefix + "test", m_test);
	if (m_test)
	{
		m_scoreFlag = farsa::ConfigurationHelper::getBool(params, prefix + "scoreFlag", m_scoreFlag);
		m_generalizationTest = farsa::ConfigurationHelper::getBool(params, prefix + "generalizationTest", m_generalizationTest);
		m_detailedTest = farsa::ConfigurationHelper::getBool(params, prefix + "detailedTest", m_detailedTest);
		m_behaviourAnalysis = farsa::ConfigurationHelper::getBool(params, prefix + "behaviourAnalysis", m_behaviourAnalysis);
		int flagCnt = 0;
		if (m_generalizationTest)
		{
			flagCnt++;
		}
		if (m_detailedTest)
		{
			flagCnt++;
		}
		if (m_behaviourAnalysis)
		{
			flagCnt++;
		}
		//if (m_generalizationTest && m_behaviourAnalysis)
		if (flagCnt > 1)
		{
			// Perform generalization test over 1000 trials
			m_generalizationTest = true;
			m_detailedTest = false;
			m_behaviourAnalysis = false;
		}
	}
	m_calcIndFit = farsa::ConfigurationHelper::getBool(params, prefix + "calcIndFit", m_calcIndFit);
	m_saveStatGenCnt = farsa::ConfigurationHelper::getInt(params, prefix + "saveBestInd", m_saveStatGenCnt);
	m_gaussianStdDev = farsa::ConfigurationHelper::getDouble(params, prefix + "gaussianStdDev", m_gaussianStdDev);
	m_defineOffspring = farsa::ConfigurationHelper::getBool(params, prefix + "defineOffspring", m_defineOffspring);
	if (m_defineOffspring)
	{
		m_numOffspring = farsa::ConfigurationHelper::getInt(params, prefix + "numOffspring", m_numOffspring);
		if (m_numOffspring == 0)
		{
			farsa::Logger::info(QString("Custom number of offspring not defined! Therefore, we use standard number of offspring, as defined in xNES"));
			m_defineOffspring = false;
		}
	}
	m_weightAnalysis = farsa::ConfigurationHelper::getBool(params, prefix + "weightAnalysis", m_weightAnalysis);
	if (m_test)
	{
		// Overwrite value read from configuration file when test flag is set
		m_weightAnalysis = false;
	}
	m_statTestGen = farsa::ConfigurationHelper::getInt(params, prefix + "statTestGen", m_statTestGen);
}

void MyEvoAlgo::describe(QString type)
{
	// Calling parent function
	farsa::Evoga::describe(type);

	Descriptor d = addTypeDescription(type, "A modified version of the steadyState algorithm that uses floating-point genes");
    d.describeReal("mutationRange").def(0.2).help("Mutation range");
	d.describeReal("stepRate").def(1.0).help("Step rate");
	d.describeInt("saveBestInd").limits(0,INT_MAX).def(0).help("How many generations both the best individual and fitness must be saved");
	d.describeReal("gaussianStdDev").def(0.1).help("Standard deviation of the Gaussian distribution");
	d.describeBool("test").def(false).help("Flags whether or not a test must be performed");
	d.describeBool("generalizationTest").def(false).help("Flags whether or not generalization test (over 1000 trials) must be run");
	d.describeBool("behaviourAnalysis").def(false).help("Flags whether or not a behavioural analysis must be run");
	d.describeBool("calcIndFit").def(false).help("Compute the individual's fitness");
	d.describeBool("scoreFlag").def(false).help("Flag that specifies whether or not the score must be saved for all replications");
}

void MyEvoAlgo::postConfigureInitialization()
{
	// Calling parent function
	farsa::Evoga::postConfigureInitialization();
	farsa::Logger::info("MyEvoAlgo Configured - Number of genes: " + QString::number(glen));
	if ((m_saveStatGenCnt <= 0) || (m_saveStatGenCnt > nogenerations))
	{
		// The best individual must be saved only at the end of the evolutionary process
		m_saveStatGenCnt = nogenerations;
	}
}

void MyEvoAlgo::evolveAllReplicas()
{
	stopEvolution = false;
	// It doesn't matter the option chosen in evolutionType,
	// this class will always call my evolutionary Algorithm.
	if (m_test)
	{
		if (m_behaviourAnalysis)
		{
			// Behavioural analysis
			analyzeBehavior();
		}
		else if (m_detailedTest)
		{
			// Detailed test (generalization test over 50 trials)
			detailedTest();
		}
		else if (m_generalizationTest)
		{
			// Generalization test (over 1000 trials)
			generalizationTest();
		}
		else
		{
			// Generalization test
			postEvaluationTest();
		}
	}
	else if (m_calcIndFit)
	{
		computeIndividualFitness();
	}
	else if (m_weightAnalysis)
	{
		analyzeWeights();
	}
	else
	{
		xnesAlgo();
	}
}

farsa::RandomGenerator* MyEvoAlgo::getRandomGenerator()
{
	return &m_rng;
}

void MyEvoAlgo::setSeed(int s)
{
	m_rng.setSeed(s);
	currentSeed = s;
}

void MyEvoAlgo::postEvaluationTest()
{
	char statfile[64];
	char genfile[64];
	char testfile[64];
	double avgFit = 0.0;
	double maxFit = -9999.0;
	double minFit = 9999.0;
	QVector<double> score(nreplications);
	for (int rp = 0; rp < nreplications; rp++)
	{
		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", this->getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) 
		{
			//sprintf(genfile,"B0S%d.gen", this->getStartingSeed() + rp);		
			sprintf(genfile, "evolvedS%d.gen", this->getStartingSeed() + rp);
			loadallg(-1,genfile);
			double fit;
			// We test the best individual of the last generation
			const int id = numLoadedGenotypes() - 1;
			exp->setNetParameters(m_pop[id]);
			exp->doAllTrialsForIndividual(id);
			fit = exp->getFitness();
			if (fit > maxFit)
			{
				maxFit = fit;
			}
			if (fit < minFit)
			{
				minFit = fit;
			}
			avgFit += fit;
			score[rp] = fit;
		} 
	}
	avgFit /= nreplications;
	// Save stats
	sprintf(testfile, "postEvaluationStats.txt");
	FILE* f = fopen(testfile, "w+");
	if (f != NULL)
	{
		fprintf(f, "Average score %lf\tMaximum score %lf\tMinimum score %lf\n", avgFit, maxFit, minFit);
		fclose(f);
	}
	if (m_scoreFlag)
	{
		char scorefile[64];
		sprintf(scorefile, "score.txt");
		FILE* fp = fopen(scorefile, "w+");
		if (fp != NULL)
		{
			for (int rp = 0; rp < nreplications; rp++)
			{
				fprintf(fp, "%lf\n", score[rp]);
			}
			fclose(fp);
		}
	}
}

void MyEvoAlgo::generalizationTest()
{
	char statfile[64];
	char genfile[64];
	char testfile[64];
	QVector<double> score(nreplications);
	double avgScore = 0.0;
	for (int rp = 0; rp < nreplications; rp++)
	{
		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", this->getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) 
		{
			//sprintf(genfile,"B0S%d.gen", this->getStartingSeed() + rp);		
			sprintf(genfile, "evolvedS%d.gen", this->getStartingSeed() + rp);			
			loadallg(-1,genfile);
			double fit;
			// We test the best individual of the last generation
			const int id = numLoadedGenotypes() - 1;
			exp->setNetParameters(m_pop[id]);
			exp->doAllTrialsForIndividual(-1);
			fit = exp->getFitness();
			score[rp] = fit;
			avgScore += fit;
		} 
	}
	avgScore /= ((double) nreplications);
	// Save stats
	sprintf(testfile, "genStats.txt");
	FILE* f = fopen(testfile, "w+");
	if (f != NULL)
	{
		fprintf(f, "Average score %lf\n", avgScore);
		fclose(f);
	}
	if (m_scoreFlag)
	{
		char scorefile[64];
		sprintf(scorefile, "genScore.txt");
		FILE* fp = fopen(scorefile, "w+");
		if (fp != NULL)
		{
			for (int rp = 0; rp < nreplications; rp++)
			{
				fprintf(fp, "%lf\n", score[rp]);
			}
			fclose(fp);
		}
	}
}

void MyEvoAlgo::detailedTest()
{
	char statfile[64];
	char genfile[64];
	char testfile[64];
	QVector<double> fit(nreplications);
	for (int rp = 0; rp < nreplications; rp++)
	{
		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", this->getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) 
		{
			sprintf(genfile, "evolvedS%d.gen", this->getStartingSeed() + rp);			
			loadallg(-1,genfile);
			// We test the best individual of the last generation
			//const int id = (nogenerations / m_saveStatGenCnt) - 1;
			const int id = numLoadedGenotypes() - 1;
			exp->setNetParameters(m_pop[id]);
			exp->doAllTrialsForIndividual(-9999 - rp);
			fit[rp] = exp->getFitness();
		} 
	}
	// Save stats
	sprintf(testfile, "detailedBehaviour.txt");
	FILE* f = fopen(testfile, "w+");
	if (f != NULL)
	{
		for (int rp = 0; rp < nreplications; rp++)
		{
			fprintf(f, "Fitness %lf\n", fit[rp]);
		}
		fclose(f);
	}
}

void MyEvoAlgo::analyzeBehavior()
{
	char statfile[64];
	char genfile[64];
	char testfile[64];
	QVector<double> fit(nreplications);
	for (int rp = 0; rp < nreplications; rp++)
	{
		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", this->getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) 
		{
			//sprintf(genfile,"B0S%d.gen", this->getStartingSeed() + rp);		
			sprintf(genfile, "evolvedS%d.gen", this->getStartingSeed() + rp);			
			loadallg(-1,genfile);
			// We test the best individual of the last generation
			const int id = numLoadedGenotypes() - 1;
			exp->setNetParameters(m_pop[id]);
			exp->doAllTrialsForIndividual(rp + 9999);
			fit[rp] += exp->getFitness();
		} 
	}
	// Save stats
	sprintf(testfile, "behaviour.txt");
	FILE* f = fopen(testfile, "w+");
	if (f != NULL)
	{
		for (int rp = 0; rp < nreplications; rp++)
		{
			fprintf(f, "Fitness %lf\n", fit[rp]);
		}
		fclose(f);
	}
}

void MyEvoAlgo::analyzeWeights()
{
	char statfile[64];
	char genfile[64];
	char testfile[64];
	QVector<double> maxw(nreplications);
	QVector<double> minw(nreplications);
	QVector<double> avgw(nreplications);
	QVector<int> outofrangew(nreplications);
	// Weight range
	farsa::ResourcesLocker locker(this);
	farsa::Evonet* evonet = getResource<farsa::Evonet>("evonet");
	const double wrange = evonet->getWrange();
	for (int rp = 0; rp < nreplications; rp++)
	{
		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", this->getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) 
		{
			sprintf(genfile, "evolvedS%d.gen", this->getStartingSeed() + rp);			
			loadallg(-1,genfile);
			// We analyze weights of the best individual of the last generation
			const int id = numLoadedGenotypes() - 1;
			if (id < 0)
			{
				farsa::Logger::error(QString("No individuals have been loaded, id = %1... Abort!").arg(id));
				exit(-1);
			}
			maxw[rp] = m_pop[id][0];
			minw[rp] = m_pop[id][0];
			avgw[rp] = m_pop[id][0];
			outofrangew[rp] = 0;
			if (fabs(m_pop[id][0]) > wrange)
			{
				outofrangew[rp]++;
			}
			for (int g = 1; g < glen; g++)
			{
				if (m_pop[id][g] > maxw[rp])
				{
					maxw[rp] = m_pop[id][g];
				}
				if (m_pop[id][g] < minw[rp])
				{
					minw[rp] = m_pop[id][g];
				}
				avgw[rp] += m_pop[id][g];
				if (fabs(m_pop[id][g]) > wrange)
				{
					outofrangew[rp]++;
				}
			}
			avgw[rp] /= ((double)glen);
		} 
	}
	// Save stats
	sprintf(testfile, "weightAnalysis.txt");
	FILE* f = fopen(testfile, "w+");
	if (f != NULL)
	{
		for (int rp = 0; rp < nreplications; rp++)
		{
			fprintf(f, "maxw %lf\tavgw %lf\tminw %lf\toutofrangew %d\n", maxw[rp], avgw[rp], minw[rp], outofrangew[rp]);
		}
		fclose(f);
	}
}

void MyEvoAlgo::computeIndividualFitness()
{
	char statfile[64];
	char genfile[64];
	char testfile[64];
	QVector<QVector<double> > fit(nreplications);
	for (int rp = 0; rp < nreplications; rp++)
	{
		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", this->getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) 
		{	
			sprintf(genfile, "individualS%d.gen", this->getStartingSeed() + rp);			
			loadallg(-1,genfile);
			const int individuals = numLoadedGenotypes();
			for (int id = 0; id < individuals; id++)
			{
				exp->setNetParameters(m_pop[id]);
				exp->doAllTrialsForIndividual(id);
				fit[rp].append(exp->getFitness() / farsa::real(exp->getNTrials()));
			}
			// Save stats
			sprintf(testfile, "individualFitnessS%d.fit", this->getStartingSeed() + rp);
			FILE* f = fopen(testfile, "w+");
			if (f != NULL)
			{
				for (int id = 0; id < individuals; id++)
				{
					fprintf(f, "%lf\n", fit[rp][id]);
				}
				fclose(f);
			}
		}
	}
}

void MyEvoAlgo::computeFStat()
{
	int i;
	double min, max, av;

	min = 9999.00;
	max = -9999.00;
	av = 0.0;

	for(i = 0; i < m_pop.size(); i++)
	{
		if((tfitness[i] / ntfitness[i]) < min)
		{
			min = tfitness[i] / ntfitness[i];
		}
		if((tfitness[i] / ntfitness[i]) > max)
		{
			max = tfitness[i] / ntfitness[i];
		}
		av += (tfitness[i] / ntfitness[i]);
	}
	this->faverage = av / (double)m_pop.size();
	this->fmax = max;
	this->fmin = min;
	this->statfit[this->cgen][0] = this->fmax;
	this->statfit[this->cgen][1] = this->faverage;
	this->statfit[this->cgen][2] = this->fmin;

	if (this->fmax > this->fbest) {
		this->fbest = this->fmax;
		this->fbestgen = this->cgen;
	}
}

void MyEvoAlgo::saveFStat()
{
	FILE *fp;
	char sbuffer[128];
	sprintf(sbuffer, "statS%d.fit", currentSeed);
    if (cgen == 0)
	{
		fp = fopen(sbuffer, "w");
	}
	else
	{
		fp = fopen(sbuffer, "a");
	}

	if (fp != NULL)
	{
		fprintf(fp, "%.3f %.3f %.3f\n", fmax, faverage, fmin);
		fclose(fp);
	}
	else
	{
		farsa::Logger::error("unable to save statistics on a file");
	}
}

void MyEvoAlgo::saveIndividual(const float* ind)
{
	char filename[64];
	FILE* fp;
	sprintf(filename, "individualS%d.gen", currentSeed);
	fp = fopen(filename, "a");
	if (fp != NULL)
	{
		fprintf(fp, "**NET : s%d_%d.wts\n", 0, 0);
		fprintf(fp, "DYNAMICAL NN\n");
		for (int j = 0; j < glen; j++)
		{
			fprintf(fp, "%lf\n", ind[j]);
		}
		fprintf(fp, "END\n");
		fclose(fp);
	}
}

void MyEvoAlgo::saveIndividualFitness(const double fit)
{
	char filename[64];
	FILE* fp;
	sprintf(filename, "individualFitnessS%d.fit", currentSeed);
	fp = fopen(filename, "a");
	if (fp != NULL)
	{
		fprintf(fp, "%lf\n", fit);
		fclose(fp);
	}
}

void MyEvoAlgo::saveEvolved(const float* ind)
{
	char filename[64];
	FILE* fp;
	sprintf(filename, "evolvedS%d.gen", currentSeed);
	fp = fopen(filename, "w");
	if (fp != NULL)
	{
		fprintf(fp, "**NET : s%d_%d.wts\n", 0, 0);
		fprintf(fp, "DYNAMICAL NN\n");
		for (int j = 0; j < glen; j++)
		{
			fprintf(fp, "%lf\n", ind[j]);
		}
		fprintf(fp, "END\n");
		fclose(fp);
	}
}

void MyEvoAlgo::saveBestInd(int bestind)
{
	char sbuffer[64];
	FILE *fp;

	sprintf(sbuffer, "B%dS%d.gen", 0, currentSeed);
	if (cgen == 0)
	{
		if ((fp = fopen(sbuffer, "w")) == NULL) 
		{
			farsa::Logger::error(QString("I cannot open file B%1S%2.gen").arg(0).arg(currentSeed));
		}
	} 
	else 
	{
		if ((fp = fopen(sbuffer, "a")) == NULL)
		{
			farsa::Logger::error(QString("I cannot open file B%1S%2.gen").arg(0).arg(currentSeed));
		}
	}

	//now saving
	fprintf(fp, "**NET : s%d_%d.wts\n", cgen, bestind);
	saveagenotype(fp, bestind);
	fflush(fp);
	fclose(fp);
}

void MyEvoAlgo::saveagenotype(FILE *fp, int ind)
{
	int j;
	fprintf(fp, "DYNAMICAL NN\n");
	for (j = 0; j < glen; j++)
	{
		fprintf(fp, "%lf\n", m_pop[ind][j]);
	}
	fprintf(fp, "END\n");
}

//save all current generation
void MyEvoAlgo::saveallg()
{
	FILE *fp;
	char filename[64];
	int i;

	sprintf(filename, "G%dS%d.gen", cgen, currentSeed);
	if ((fp = fopen(filename, "w+")) == NULL) 
	{
		farsa::Logger::error(QString("Cannot open file %1").arg(filename));
	} 
	else 
	{
		//we save
		for(i = 0; i < popSize;i++) 
		{
			fprintf(fp, "**NET : %d_%d_%d.wts\n", cgen, 0, i);
			saveagenotype(fp, i);
		}
		fclose(fp);
	}
}

void MyEvoAlgo::loadgenotype(FILE *fp, int ind)
{
	int j;
	double v;

	fscanf(fp, "DYNAMICAL NN\n");
	for (j = 0; j < glen; j++) 
	{
		fscanf(fp, "%lf\n", &v);
		m_pop[ind][j] = v;
	}
	fscanf(fp, "END\n");
}

int MyEvoAlgo::loadallg(int gen, const char *filew)
{
	FILE *fp;
	char filename[512];
	char message[512];
	char flag[512];

	if (gen >= 0) 
	{
		sprintf(filename, "G%dS%d.gen", gen, seed);
	} 
	else 
	{
		sprintf(filename, "%s", filew);//in case of B%P$S.gen
	}

	if ((fp = fopen(filename, "r")) != NULL) 
	{
		m_pop.clear();
		int idx = 0;
		while (true) 
		{
			flag[0] = '\0';
			fscanf(fp, "%s : %s\n", flag, message);
			if (strcmp(flag, "**NET") == 0)
			{
				float* ind = new float[glen];
				m_pop.append(ind);
				loadgenotype(fp, idx);
				idx++;
			} 
			else 
			{
				break;
			}
		}
		farsa::Logger::info(QString("Loaded ind: %1").arg(m_pop.size()));
		fclose(fp);
	} 
	else 
	{
		farsa::Logger::error(QString("File %1 could not be opened").arg(filename));
	}

	loadedIndividuals = m_pop.size();

	return loadedIndividuals;
}

unsigned int MyEvoAlgo::loadGenotypes(QString filename)
{
	return loadallg(-1, filename.toLatin1().data());
}

unsigned int MyEvoAlgo::numLoadedGenotypes() const
{
	return loadedIndividuals;
}

int MyEvoAlgo::getLoadedIndividualLength(const int id)
{
	return glen;
}

int MyEvoAlgo::round(double value)
{
	int val;
	// Compute the nearest integer less than value
	int lowBound = floor(value);
	// Compute the decimal part
	double diff = value - (lowBound);
	if (diff >= 0.5)
	{
		// If decimal part is >= 0.5, round to the next integer
		val = (lowBound + 1);
	}
	else
	{
		// Otherwise, set val to the lower bound
		val = lowBound;
	}
	return val;
}

/*
 * Main function of the xNES algorithm (exponential Natural Evolution Strategies)
 */
void MyEvoAlgo::xnesAlgo()
{
	// ALL VARIABLES ARE INITIALIZED TO AVOID NUMERICAL PROBLEMS

	int rp; //replication
	int gn; //generation
	const float genLen = (const float)glen; // Floating point genome length
	float offspringSize = 4.0 + floor(3.0 * log(genLen)); // Floating point number of offspring generated by the individual
	int offSize = (int)floor(offspringSize); // Number of offspring generated by the individual
	if (m_defineOffspring)
	{
		// Before setting, we perform a dummy check (i.e., we already checked this during configuration phase)
		if (m_numOffspring != 0)
		{
			offSize = m_numOffspring;
		}
	}
	Eigen::VectorXf ind(glen); // The individual from which the offspring will be generated
	Eigen::VectorXf dInd(glen); // The update of the individual
	Eigen::MatrixXf covMatrix(glen,glen); // Covariance Matrix
	Eigen::MatrixXf dCovM(glen,glen); // The update of the covariance matrix
	Eigen::MatrixXf expCovM(glen,glen); // Exponential matrix of CovM
	float etaInd; // Learning rate for the individual
	float etaCovM; // Learning rate for the covariance matrix
	Eigen::VectorXf utility(offSize); // Array containing the utility values
	float utilSum; // Sum of utility values
	float tmp; // Temporary variable
	Eigen::VectorXf orderedUtility(offSize); // Array containing the ordered utilities (i.e. the best individual has the highest utility value)
	float ordUtilSum; // Sum of ordered utility values (it shall be equal to 1.0)
	QVector<float> f(offSize); // Array containing the fitness of the offspring
	Eigen::VectorXf ordf(offSize); // Array containing the ordered fitness of the offspring
	QVector<int> ordoff(offSize); // Array containing the ordered indices of the offspring
	float* currInd = new float[glen]; // Current individual to be evaluated (temporary variable)
	Eigen::MatrixXf eye(glen,glen); // Identity matrix
	// Indices
	int i = 0;
	int j = 0;
	// Start generation index
	int startGeneration;
	// File descriptors
	char statfile[128];
	char genFile[128];
	// Weight range
	farsa::ResourcesLocker locker(this);
	farsa::Evonet* evonet = getResource<farsa::Evonet>("evonet");
	const double wrange = evonet->getWrange();

	// Resize population
	m_pop.resize(offSize);
	for (int i = 0; i < offSize; i++)
	{
		m_pop[i] = new float[glen];
	}

	farsa::Logger::info("EVOLUTION: xNES");
	farsa::Logger::info("Number of replications: " + QString::number(nreplications));

	startGeneration = 0;

	// Initialize XNES parameters
	etaInd = 1.0;
	etaCovM = (9.0 + 3.0 * log(genLen)) / (5.0 * genLen * sqrt(genLen));
	// Calculate the utility values before normalization
	utilSum = 0.0;
	for (j = 0; j < offSize; j++)
	{
		// u[j] is equal to the uttermost between 0 and (log(lambda / 2 + 1) - log(j))
		tmp = log(offspringSize / 2.0 + 1.0) - log((float)j + 1.0); // Remember that log(0) = -Inf, so 1 is added to index i
		if (tmp < 0.0)
		{
			tmp = 0.0;
		}
		utility[j] = tmp;
		// Update the sum of utility values
		utilSum += utility[j];
	}
	// Normalize the utility values
	for (j = 0; j < offSize; j++)
	{
		if (utilSum != 0.0)
		{
			utility[j] /= utilSum;
		}
		utility[j] -= 1.0 / genLen; // In the standard version, utility values might be negative
	}

	for (i = 0; i < glen; i++)
	{
		// Define the identity matrix
		eye(i,i) = 1.0;
		for (j = 0; j < glen; j++)
		{
			if (i != j)
			{
				eye(i,j) = 0.0;
			}
			// Initialize covariance matrix, its update and its exponential
			covMatrix(i,j) = 0.0;
			dCovM(i,j) = 0.0;
			expCovM(i,j) = 0.0;
		}
		// Initialize individual and its update
		ind[i] = 0.0;
		dInd[i] = 0.0;
		// Initialize the individual (temporary variable) used for evaluation
		currInd[i] = 0.0;
	}

	for(rp = 0; rp < nreplications; rp++) 
	{ // replications
		startGeneration = 0;
		setSeed(getStartingSeed() + rp);
		farsa::Logger::info(QString("Replication %1, seed: %2").arg(rp + 1).arg(getStartingSeed() + rp));
		resetGenerationCounter();

		// Set fbest to a very low value
		this->fbest = -99999.0;

		QTime evotimer;
		evotimer.start();
		// Initialization of fitnesses
		for (j = 0; j < offSize; j++) {
			tfitness[j] = 0.0;
			ntfitness[j] = 0.0;
		}

		//code to recovery a previous evolution: Experimental
		sprintf(statfile,"statS%d.fit", getStartingSeed() + rp);
		//now check if the file exists
		farsa::DataChunk statTest(QString("stattest"),Qt::blue,2000,false);
		if (statTest.loadRawData(QString(statfile),0)) {
			startGeneration=statTest.getIndex();
			sprintf(genFile,"G%dS%d.gen", startGeneration, getStartingSeed() + rp);
			farsa::Logger::info("Recovering from startGeneration: " + QString::number(startGeneration));
			farsa::Logger::info(QString("Loading file: ") + genFile);
			loadallg(-1, genFile);
			cgen = startGeneration;
			emit recoveredInterruptedEvolution( QString(statfile) );
		} //end evolution recovery code

		// Initialize the individual for this replication
		for (i = 0; i < glen; i++)
		{
			//ind[i] = pool[poolBest][i];
			ind[i] = (getRandomGenerator())->getDouble(-wrange, wrange);
			// Reset the current offspring (to avoid numerical issues)
			currInd[i] = 0.0;
		}

		// Initialize Covariance Matrix
		for (i = 0; i < glen; i++)
		{
			for (j = 0; j < glen; j++)
			{
				covMatrix(i,j) = 0.0;
				expCovM(i,j) = 0.0;
			}
		}

		// Initialize fitnesses, ordered offspring and ordered utilities
		for (j = 0; j < offSize; j++) // Remember that the last slot is reserved for the parent
		{
			f[j] = 0.0;
			ordf[j] = 0.0;
			ordoff[j] = 0.0;
			orderedUtility[j] = 0.0;
		}
		
		for(gn = startGeneration; gn < nogenerations; gn++) 
		{	// generations
			evotimer.restart();
			farsa::Logger::info("Generation " + QString::number(gn));

			// Calculate the matrix exponential
			Eigen::MatrixExponential<Eigen::MatrixXf>(covMatrix).compute(expCovM);

			// Extracting samples (i.e. generate offspring of current individual)
			Eigen::MatrixXf samples(glen,offSize);
			Eigen::MatrixXf trSamples(glen,offSize);
			Eigen::MatrixXf offspring(glen,offSize);

			for (i = 0; i < glen; i++)
			{
				for (j = 0; j < offSize; j++)
				{
					samples(i,j) = (getRandomGenerator())->getGaussian(m_gaussianStdDev, 0.0);
				}
			}

			exp->initGeneration(gn);

			// Create the offspring
			trSamples = expCovM * samples;
			for (j = 0; j < offSize; j++)
			{
				for (i = 0; i < glen; i++)
				{
					offspring(i,j) = ind[i] + trSamples(i,j);
				}
			}

			// Fill the current offspring
			for (j = 0; j < offSize; j++)
			{
				for (i = 0; i < glen; i++)
				{
					currInd[i] = offspring(i,j);
					// Save offspring genotypes
					m_pop[j][i] = currInd[i];
				}
				// Set the ANN parameters
				exp->setNetParameters(currInd);
				// Perform trials for the individual (i.e. the offspring)
				exp->doAllTrialsForIndividual(j);
				// Calculate the fitness of the individual (i.e. the offspring)
				f[j] = exp->getFitness();

				tfitness[j] = f[j];
				ntfitness[j] = 1.0;

				if (isStopped()) 
				{ // stop evolution
					return;
				}
			}

			// Now fitnesses have to be ordered in descending order
			// (in the original XNES, fitnesses are ordered in ascending order)
			int startIndex = 0;
			QVector<bool> examined(offSize);
			for (j = 0; j < offSize; j++)
			{
				examined[j] = false;
			}
			int countExamined = 0;
			while (countExamined < offSize)
			{
				bool found = false;
				i = 0;
				float maxf;
				int indMaxf;
				while (i < offSize && !found)
				{
					if (!examined[i])
					{
						maxf = f[i];
						indMaxf = i;
						found = true;
					}
					i++;
				}
				for (j = 0; j < offSize; j++)
				{
					if (f[j] > maxf && !examined[j])
					{
						maxf = f[j];
						indMaxf = j;
					}
				}
				ordf[startIndex] = maxf;
				ordoff[startIndex] = indMaxf;
				examined[indMaxf] = true;
				countExamined++;
				startIndex++;
			}

			// Initialize the sum of ordered utility values
			ordUtilSum = 0.0;
			// Now fill the array with the utility values
			for (j = 0; j < offSize; j++)
			{
				orderedUtility[ordoff[j]] = utility[j];
				ordUtilSum += orderedUtility[ordoff[j]];
			}

			// Compute the natural gradient for both the individual and the covariance matrix

			// Create a matrix whose rows contain weights * samples
			Eigen::MatrixXf ordUtil(glen,offSize);
			Eigen::MatrixXf ordUtilSamples(glen,offSize);
			Eigen::MatrixXf Z(offSize,glen);

			for (i = 0; i < glen; i++)
			{
				for (j = 0; j < offSize; j++)
				{
					ordUtil(i,j) = orderedUtility[j];
				}
			}

			for (i = 0; i < glen; i++)
			{
				for (j = 0; j < offSize; j++)
				{
					ordUtilSamples(i,j) = ordUtil(i,j) * samples(i,j);
					Z(j,i) = samples(i,j);
				}
			}

			Eigen::MatrixXf G = ordUtilSamples * Z - ordUtilSum * eye;

			// Define transpose matrices for both samples and ordered utilities
			// (they are necessary to perform the calculation of gradients)
			Eigen::MatrixXf ordUtilTr(offSize,1);
			Eigen::MatrixXf v(glen,1);
			for (j = 0; j < offSize; j++)
			{
				ordUtilTr(j,0) = orderedUtility[j];
			}

			v = etaInd * expCovM * samples * ordUtilTr;
			for (i = 0; i < glen; i++)
			{
				dInd[i] = v(i,0);
			}
			dCovM = etaCovM * G;
			// Update the individual and the covariance matrix
			for (i = 0; i < glen; i++)
			{
				ind[i] += (m_stepRate * dInd[i]);
			}
			covMatrix += (m_stepRate * dCovM);

			// Evaluate the individual
			double indfit = 0.0;
			for (i = 0; i < glen; i++)
			{
				currInd[i] = ind[i];
			}
			// Set the ANN parameters
			exp->setNetParameters(currInd);
			// Perform trials for the individual (i.e. the offspring)
			exp->doAllTrialsForIndividual(0);
			// Calculate the fitness of the individual (i.e. the offspring)
			indfit = exp->getFitness();

			exp->endGeneration(gn);

			if ( commitStep() ) { return; }

			// Compute and save statistics
			computeFStat();
			// Save the best individual (in terms of its fitness) among the offspring
			//if ((gn == 0) || (((gn + 1) % m_saveStatGenCnt) == 0))
			//{
				saveBestInd(ordoff[0]);
				saveFStat();
				// Save current individual too
				saveIndividual(currInd);
				saveIndividualFitness(indfit);
			//}

			emit endGeneration( cgen, fmax, faverage, fmin );
			if (commitStep()) {
				return; // stop the evolution process
			}

			cgen++;

			// Reset weights and fitnesses once used
			for (j = 0; j < offSize; j++)
			{
				f[j] = 0.0;
				orderedUtility[j] = 0.0;
				tfitness[j] = 0.0;
				ntfitness[j] = 0;
			}
			// Reset matrix exponential
			for (i = 0; i < glen; i++)
			{
				for (j = 0; j < glen; j++)
				{
					expCovM(i,j) = 0.0;
				}
			}

			// Reset current offspring
			for (i = 0; i < glen; i++)
			{
				currInd[i] = 0.0;
			}

			// Print best fitness found at current generation
            farsa::Logger::info(QString("Generation %1 Time %2m  Best Fitness %3").arg(gn+1).arg((double)evotimer.elapsed()/60000.0, 0, 'f', 2).arg(fmax));
			fflush(stdout);
		}
		// Save the best generation fitness statistics
		saveBestFitness();
		// Save the individual at the end of evolutionary process
		float* evolvedInd = new float[glen];
		for (i = 0; i < glen; i++)
		{
			evolvedInd[i] = ind[i];
		}
		saveEvolved(evolvedInd);
	}
}
